##############################################
#                                            #
# TARV: Tree-based Analysis of Rare Variants #
# Chi Song and Heping Zhang                  #
#                                            #
# Code by Chi Song                           #
#                                            #
##############################################


## Function to calculate the marginal effect of each rare variant
##   Arguments:
##     pheno    - A data.frame of the phenotype data, including the disease response, each row is a sample
##     geno     - A matrix of the genotypes of each rare variant for all the samples, each column is a sample, and each row is a variant
##     form.adj - A character string of the formula to adjust for the effect of clinical confounders, e.g. "disease ~ sex + race"
##     family   - The family of the regression, "binomial" for logistic regression and "gaussian" for linear regression

TARV.marginal.effect <- function(pheno, geno, form.adj, family = c("binomial", "gaussian"))
{
    family <- match.arg(family)
    if(nrow(pheno) != ncol(geno)) stop("Error: nrow(pheno) != ncol(geno)\n")
    form <- as.formula(paste(as.character(form.adj), "+ geno[i, ]"))
    Z <- rep(NA, nrow(geno))
    for(i in 1:nrow(geno)) {
        model <- glm(form, data = pheno, family = family)
        Z[i] <- tail(summary(model)$coefficient, n = 1)[3]
    }
    return(Z)
}


## Function to transform the genotype data to a new matrix
##   Arguments:
##     geno       - Genotypes of the rare variants, the same as geno in TARV.marginal.effect
##     annotation - A vector of the anootation information of each variant, usually the gene name
##     Z          - The statistic used to rank the variants, by default the output from TARV.marginal.effect
##     direction  - The direction of the effect of interest

TARV.transform <- function(geno, annotation, Z, direction = c("both", "positive", "negative"))
{
    direction <- match.arg(direction)
    if(nrow(geno) != length(annotation)) stop("Error: nrow(geno) != length(annotation)\n")
    if(nrow(geno) != length(Z)) stop("Error: nrow(geno) != length(Z)\n")
    gid <- unique(annotation)
    if(direction %in% c("positive", "both")) {
        X.pos <- matrix(NA, ncol(geno), length(gid))
        colnames(X.pos) <- paste(gid, "+", sep = "")
    }
    if(direction %in% c("negative", "both")) {
        X.neg <- matrix(NA, ncol(geno), length(gid))
        colnames(X.neg) <- paste(gid, "-", sep = "")
    }
    for(i in 1:length(gid)) {
        g <- gid[i]
        gindex <- annotation == g
        z <- Z[gindex]
        if(direction %in% c("positive", "both")) {
            order.pos <- order(-z)
            index.pos <- which(gindex)[order.pos]
            X.pos[, i] <- apply(rbind(geno[index.pos, ] > 0, TRUE), 2, function(x) which(x)[1])
        }
        if(direction %in% c("negative", "both")) {
            order.neg <- order(z)
            index.neg <- which(gindex)[order.neg]
            X.neg[, i] <- apply(rbind(geno[index.neg, ] > 0, TRUE), 2, function(x) which(x)[1])
        }
    }
    if(direction == "both") {
        X <- cbind(X.pos, X.neg)
    } else if(direction == "positive") {
        X <- X.pos
    } else if(direction == "negative") {
        X <- X.neg
    }
    return(X)
}


## Function to output the tranformed data for RTREE usage
##   Arguments:
##     pheno    - Phenotype data, the same as geno in TARV.marginal.effect
##     response - The column number or name of the disease response in pheno
##     delete   - Optional, the column number or name of pheno variable to be excluded from the RTREE analysis
##     X        - The transformed genotype data, by default the output form TARV.transform
##     basename - The basename of the output files, two output files will be generated "basename_data.txt" for the input of RTREE program, and "basename_varlist.txt" for the list of variables used in RTREE

TARV.output.rtree <- function(pheno, response, delete = NULL, X, basename)
{
    if(nrow(pheno) != nrow(X)) stop("Error: nrow(pheno) != nrow(X)\n")
    if(length(response) > 1) stop("Error: length(response) > 1\n")
    if(is.numeric(response)) {
        if(max(response) > ncol(pheno)) stop ("Error: response > ncol(pheno)\n")
    } else {
        if(!(response %in% colnames(pheno))) stop("Error: response not found in colnames(pheno)\n")
    }
    if(!is.null(delete)) {
        if(is.numeric(delete)) {
            if(max(delete) > ncol(pheno)) stop ("Error: delete > ncol(pheno)\n")
        } else {
            if(!(delete %in% colnames(pheno))) stop("Error: delete not found in colnames(pheno)\n")
        }
    }
    data <- cbind(pheno, X)
    type <- rep("o", ncol(data))
    type[which(sapply(pheno, class) == "factor")] <- "n"
    if(is.numeric(response)) type[response] <- "r" else type[which(colnames(pheno) == response)] <- "r"
    if(!is.null(delete)) if(is.numeric(delete)) type[delete] <- "d" else type[which(colnames(pheno) == delete)] <- "d"
    var.num <- paste("X", 1:sum(!(type %in% c("r", "d"))), sep = "")
    var.name <- colnames(data)[!(type %in% c("r", "d"))]
    varlist <- data.frame(variable = var.num, name = var.name)
    file.data <- paste(basename, "_data", ".txt", sep = "")
    file.var <- paste(basename, "_varlist", ".txt", sep = "")
    writeLines(c(paste(type, collapse = "\t"), apply(data, 1, paste, collapse = "\t")), con = file.data)
    write.table(varlist, quote = FALSE, sep = "\t", na = "", row.names = FALSE, col.names = TRUE, file = file.var)
}


## Example code
## Simulate a small case-control dataset with 5 genes (20 rare variants in each gene)
sex <- rbinom(1000, 1, 0.5)
race <- sample(1:3, 1000, replace = TRUE, prob = c(0.4, 0.4, 0.2))
gg <- replicate(100, rbinom(1000, 1, runif(1, 0.005, 0.05)))
annotation <- paste("gene", rep(1:5, each = 20), sep = "")
causal <- rbinom(40, 1, 0.8)
x1 <- rowSums(gg[, 1:20][, causal[1:20] > 0]) > 0
x2 <- rowSums(gg[, 21:40][, causal[21:40] > 0]) > 0
xb <- sex * 0.2 + (race == 2) * 0.2 + x1 * 0.6 + x2 * 0.8 - 0.7
r <- rbinom(1000, 1, exp(xb)/(1+exp(xb)))
table(r)
pheno <- data.frame(disease = r, sex = sex, race = as.factor(race))
geno <- t(gg)

## Calculate the marginal effects using Logistic regression
Z <- TARV.marginal.effect(pheno, geno, "disease ~ sex + race", family = "binomial")
## Transform data
X <- TARV.transform(geno, annotation, Z, direction = "both")
## Output data for RTREE
TARV.output.rtree(pheno, 1, NULL, X, "demo")
